/*********************************************************************************
*        File: ALV2
*        Controller: PIC16f1847
*        Compiler: Hitech C
*        Date: 16.04.2023
*        Author: Axel Joost
*        Description: Wechselrichter oder Akkus ber PWM steuern, bzw. laden
************************************************************************************
         Header files
************************************************************************************/
#include <pic.h>      
#include <pic16f1847.h>        
/***********************************************************************************
         Configuration Bits
************************************************************************************/
__CONFIG(FOSC_INTOSC & PWRTE_ON & BOREN_ON & MCLRE_OFF & WDTE_OFF & CP_OFF);
__CONFIG (LVP_OFF); // Config 1 + 2
/***********************************************************************************
         Konstanten
************************************************************************************/
#define  T_UP     RA3
#define  T_MODE   RA5
#define  T_DOWN   RA2
#define  RELAIS   LATB3
#define  LCD_EN   LATA7
#define  LCD_RS   LATA6
#define  LADE     RB1
#define  ENTLADE  RB2
#define  EEDATA   EEDATL
/**********************************************************************************
         Variablen, Programmdeklarationen
***********************************************************************************/           
unsigned char TEMP, ARBEITSFLAG, SCHALTFLAG, LADEFLAG, SHELLY, LADE_Z, INV_Z, PWM_MIN;
unsigned int TEMP1, DATA1, P_IST, P_SOLL, P_ALT, U_IN, U_MIN, U_MAX, I_INV, U_OUT;
void Init (void);            
void Delaymittel(void);           
void LCDDelay4ms(void);
void LCDDelay15ms (void);
void LCDDelay500ns (void);    
void LCDWrite (char TEMP);
void LCDText (const char *buffer);
void LCDInit (void);
void LCDZ1 (void);
void LCDZ2 (void);
void LCDAD1(void);
void LCDAD2(void);
void LCDLeer (void);
void EESchreibe (void);          
void EELese (void);     
void Taste (void);
void AD (void);
void Steuerung (void);
void Mpp(void);
void Rauf(void);
void Runter (void);


/**********************************************************************************
         Main
***********************************************************************************/
void main (void)                  												
{ 
  Init();
  LCDInit();
  RELAIS = 1;
  LCDZ1();
  LCDText("SolarElectriX.de");
  TEMP = 0;
  LADE_Z = 0;
  INV_Z = 0;
  P_ALT = 0;
  Delaymittel();
  CCPR1L = PWM_MIN;
  while (1)
  {
    U_MAX = U_MIN;                                                              // U_MAX wird zum Laden und U_MIN zum Entladen genutzt. Das ging besser in meinen Kopf rein..
    if (T_MODE == 0)
    Taste();
    AD();
    if (ARBEITSFLAG == 1)                                                       // MPP Laderegler
    Mpp();
    else                                                                          
    Steuerung();                                                                // Laden ber Shelly und Entladen ber Shelly oder mit Festwerten
    Delaymittel();
  }
}


/**********************************************************************************
         Unterprograme
***********************************************************************************/
void Steuerung (void)                                                           
{
  if (TEMP < 25)
  TEMP++;
  if (P_IST > P_SOLL + 100)                                                     // Wenn viel zu hohe Leistung, dann sofort runterregeln und nicht erst warten, bis der FET hei ist. 
  TEMP = 25;                                                                    
  if (TEMP == 25)                        										// Ca. 1 Sekunde, kann vielleicht etwas schneller, wenn die Shelly - Cloud mitspielt...
  {      
    if ((LADE == 0)&&(INV_Z < (181 - PWM_MIN))&&(LADE_Z == 0))                  // LADE_Z und INV_Z sind Zhler, die immer mitlaufen, damit ein Sun 2 Grid, der z.B. zum Laden
    INV_Z++;                                                                    // genutzt wird, auch "wei", wann ein anderer Sun 2 Grid, der zum Entladen genutzt wird, sein PWM 
    if (( LADE == 1)&&(INV_Z > 0))                                              // runtergefahren hat und dann erst eigenes PWM erst hochfhrt, wenn das andere Modul fertig ist.
    INV_Z--;                                                                    // Ansonsten wrden die Module u.U. gleichzeitig arbeiten, also ein Modul wrde laden und das andere
                                                                                // gleichzeitig entladen und das wre bld...    
    if ((ENTLADE == 0)&&(LADE_Z < (181 - PWM_MIN))&&(INV_Z == 0))
    LADE_Z++;
    if ((ENTLADE == 1)&&(LADE_Z > 0))
    LADE_Z--;
    
    if ((LADEFLAG == 0)&&(SHELLY == 1))  										// Batterie bedient den Wechselrichter, ber Shelly regelbar
    {
      P_SOLL = 500;
      if (((CCPR1L > PWM_MIN)&&(ENTLADE == 1))||(U_IN < U_MIN+3))
      Runter();

      if ((P_IST < P_SOLL)&&(CCPR1L < 255)&&(U_IN > U_MIN)&&(ENTLADE == 0)&&(INV_Z == 0))
      Rauf();
    }

    if ((LADEFLAG == 0)&&(SHELLY == 0))  										// Batterie bedient den Wechselrichter mit Festwerten
    {
      if (((P_IST > P_SOLL)&&(CCPR1L > PWM_MIN))||(U_IN < U_MIN+3))
      Runter();

      if ((P_IST < P_SOLL)&&(CCPR1L < 255)&&(U_IN > U_MIN+2))
      Rauf();
    }

    if (LADEFLAG == 1)  			                  							// Netzteil ladet die Batterie, ber Shelly regelbar
    {
      if (((CCPR1L > PWM_MIN)&&(LADE == 1))||(U_OUT > U_MAX-2))
      Runter();

      if ((P_IST < P_SOLL)&&(CCPR1L < 255)&&(LADE == 0)&&(U_OUT < U_MAX -1)&&(LADE_Z == 0)) 
      Rauf();
    }

    TEMP = 0;
    LCDZ1();
    if (LADEFLAG == 0)
    LCDText("Inv.  Soll   PWM");  
    if (LADEFLAG == 1)
    LCDText("Lad.  U_Bat. PWM");
    LCDZ2();
    if (P_IST > 1)
    DATA1 = P_IST; 
    else DATA1 = 0;   
    LCDAD2();
    if ((LADEFLAG == 0)&&(SHELLY == 0))
    {
      LCDText("W  ");
      DATA1 = P_SOLL;
      LCDAD2();
      LCDText("W   ");
    }
    if ((LADEFLAG == 0)&&(SHELLY == 1))
    {
      LCDText("W  Var.   ");
    }
    if (LADEFLAG == 1)
    {
      LCDText("W ");
      DATA1 = U_OUT;
      LCDAD1();
      LCDText("V  ");
    }
  /* Zum Testen von INV_Z und LADE_Z:
    DATA1 = INV_Z;
    LCDAD2();
    LCDText("In.");
    DATA1 = LADE_Z;
    LCDAD2();
    LCDText("La.");*/

    DATA1 = CCPR1L;                                                              // Ist U_MAX, wenn geladen wird.
    LCDAD2();
  } 
}                                        										// Ende Steuerung


/* Anmerkung: Die folgenden 2 Routinen kann man noch bestimmt noch optimieren, um die Reaktionszeiten zu verkleinern.
   Vielleicht macht es Sinn, fr Ladung und Entladung verschiedene Routinen zu haben ?
   Im Moment sind die Routinen fr Ladung und Entladung gleich. Diese knnte man auch an verschiedene Wechselrichter anpassen,
   da diese z.B. bei verschiedenen unteren PWM - Werten anfangen "zu leben". ... Einfach mal mit rumspielen :-)
*/

void Rauf (void)                                                            
{
//  if (LADEFLAG == 0)                                                            // Entladung
  {                                                 
    if (CCPR1L < (PWM_MIN + 40))
    CCPR1L = (CCPR1L + 5);
    if ((CCPR1L > (PWM_MIN + 39))&&(CCPR1L < (PWM_MIN + 80)))
    CCPR1L = (CCPR1L + 3);
    if (CCPR1L > (PWM_MIN + 79))
    CCPR1L++;
  }

/*  if (LADEFLAG == 1)                                                             // Ladung
  {
    if (CCPR1L < (PWM_MIN + 80))
    CCPR1L = (CCPR1L + 10);
    if (CCPR1L > (PWM_MIN + 79))
    CCPR1L++;
  }*/
}


void Runter (void)
{
 // if (LADEFLAG == 0)
  {
    if ((CCPR1L > PWM_MIN)&&(CCPR1L < (PWM_MIN + 47)))
    CCPR1L = (CCPR1L - 5);
    if ((CCPR1L > (PWM_MIN + 46))&&(CCPR1L < (PWM_MIN + 85)))
    CCPR1L = (CCPR1L - 3);
    if (CCPR1L > (PWM_MIN + 84))
    CCPR1L--;
    if ((CCPR1L < PWM_MIN)&&(U_IN > U_MIN))
    CCPR1L = PWM_MIN;
    else
    if (U_IN < U_MIN)
    CCPR1L = 0;
  }

 /* if (LADEFLAG == 1)
  {
    if (CCPR1L < (PWM_MIN + 80))
    CCPR1L = (CCPR1L - 10);
    if (CCPR1L > (PWM_MIN + 79))
    CCPR1L--;
    if (CCPR1L < (PWM_MIN + 9))
    CCPR1L = PWM_MIN;
  }*/
}


void Mpp (void)                                                                 // MPP Lader vom Panel in Akku, muss noch getestet werden...
{
  P_SOLL == 400;                                                                   
  if (TEMP < 5)
  TEMP++;
  if (TEMP > 4)
  {                                     										// Ab hier Anzeige der Werte
    LCDZ1();
    LCDText("MPP - Laderegler");  
    LCDZ2();
    if (P_IST > 1)
    DATA1 = P_IST; 
    else DATA1 = 0;   
    LCDAD2();
    LCDText(" W   PWM ");
    DATA1 = CCPR1L;                                                              // Ist U_MAX, wenn geladen wird.
    LCDAD2();
    TEMP = 0;
  }
  Delaymittel();
  AD();
  Delaymittel();
  if ((P_IST > P_ALT)&&(CCPR1L < 255)&&(P_IST < P_SOLL))
  CCPR1L++;
  if ((P_IST == P_ALT)&&(CCPR1L < 255))
  CCPR1L++;
  if (((P_IST < P_ALT)&&(CCPR1L > 0))||(P_IST > P_SOLL)||(U_OUT > U_MAX))
  CCPR1L--;
  P_ALT = P_IST;                                                                 // Neuer Wert wird alter Wert
}


void AD (void)                                                                   // Spannungs - und Strommessungen, Vref. 4,096 Volt 
{
  ADCON0 = 0b00000011;                           // AD0
  while (GO_nDONE);
  U_IN = ((ADRESH<<8|ADRESL)+ U_IN)*0.39;
  ADCON0 = 0b00000111;                           // AD1
  while (GO_nDONE);
  I_INV = ((ADRESH<<8|ADRESL)+ I_INV)*0.5;
  ADCON0 = 0b00010011;                           // AD4
  while (GO_nDONE);
  U_OUT = ((ADRESH<<8|ADRESL)+ U_OUT)*0.39;
  P_IST = 0.0150 * ((U_IN/3) * (I_INV/2));
}


void Taste (void)                                                                // Menue zum Einstellen der Funktionen und Parameter
{  
  CCPR1L = 0;
  while (T_MODE == 0)
  Delaymittel();
  LCDLeer();
  LCDZ1();
  LCDText("Einstellungen   ");  
  while (T_MODE == 1)
  {
  
    if (T_DOWN == 0)
    { 
      while (T_DOWN == 0)
      Delaymittel();
      if (ARBEITSFLAG > 0)
      ARBEITSFLAG--;
      if (ARBEITSFLAG == 0)
      ARBEITSFLAG = 4;
    }

    if (T_UP == 0)
    {
      while (T_UP == 0)
      Delaymittel();
      if (ARBEITSFLAG < 5)
      ARBEITSFLAG++;
      if (ARBEITSFLAG == 5)
      ARBEITSFLAG = 1;
    }

    LCDZ2();
    if (ARBEITSFLAG == 1)
    LCDText("MPP - Laderegler");
    if (ARBEITSFLAG == 2)
    LCDText("Bat. -> Inverter");
    if (ARBEITSFLAG == 3)
    LCDText("Netztteil -> Bat");
    if (ARBEITSFLAG == 4)
    LCDText("Shelly aktiv ?  ");
    Delaymittel();
  }
  while (T_MODE == 0)
  Delaymittel();

  if (ARBEITSFLAG == 1)                          // MPP Laderegler
  {
    while (T_MODE == 1)
    {
      LCDZ2();
      LCDText("Bat.-Max./V");
      if ((T_UP == 0)&&(U_MIN < 550))
      U_MIN++;
      if ((T_DOWN == 0)&&(U_MIN > 200))
      U_MIN--;
      DATA1 = U_MIN;
      LCDAD1(); 
      Delaymittel();
    }
    while (T_MODE == 0)
    Delaymittel();
    CCPR1L = 255;
  }

  if (ARBEITSFLAG == 2)                         // Bat. -> Inverter
  {
    while (T_MODE == 1)
    {
      LCDZ2();
      LCDText("Leistung/W ");
      if ((T_UP == 0)&&(P_SOLL < 500))
      P_SOLL++;
      if ((T_DOWN == 0)&&(P_SOLL > 0))
      P_SOLL--;
      DATA1 = P_SOLL;
      LCDAD2(); 
      LCDText("  ");
      Delaymittel();
    }
    while (T_MODE == 0)
    Delaymittel();
    while (T_MODE ==1)
    {
      LCDZ2();
      LCDText("Bat.-Min./V");
      if ((T_UP == 0)&&(U_MIN < 550))
      U_MIN++;
      if ((T_DOWN == 0)&&(U_MIN > 200))
      U_MIN--;
      DATA1 = U_MIN;
      LCDAD1(); 
      Delaymittel();
    }
    while (T_MODE == 0)
    Delaymittel();
    while (T_MODE ==1)
    {
      LCDZ2();
      LCDText("PWM - Min.   ");
      if ((T_UP == 0)&&(PWM_MIN < 120))
      PWM_MIN++;
      if ((T_DOWN == 0)&&(PWM_MIN > 0))
      PWM_MIN--;
      DATA1 = PWM_MIN;
      LCDAD2(); 
      Delaymittel();
    }
    while (T_MODE == 0)
    Delaymittel();
    LADEFLAG = 0;
    CCPR1L = PWM_MIN;
  }
  
  if (ARBEITSFLAG == 3)                                  // Netzteil -> Bat.
  {
    while (T_MODE ==1)
    {
      LCDZ2();
      LCDText("Bat.-MAX./V");
      if ((T_UP == 0)&&(U_MIN < 550))
      U_MIN++;
      if ((T_DOWN == 0)&&(U_MIN > 200))
      U_MIN--;
      DATA1 = U_MIN;
      LCDAD1(); 
      Delaymittel();
    }
    while (T_MODE == 0)
    Delaymittel();
    while (T_MODE ==1)
    {
      LCDZ2();
      LCDText("PWM - Min.   ");
      if ((T_UP == 0)&&(PWM_MIN < 120))
      PWM_MIN++;
      if ((T_DOWN == 0)&&(PWM_MIN > 0))
      PWM_MIN--;
      DATA1 = PWM_MIN;
      LCDAD2(); 
      Delaymittel();
    }
    while (T_MODE == 0)
    Delaymittel();
    LADEFLAG = 1;
    SHELLY = 1;
    CCPR1L = PWM_MIN;
  }

  if (ARBEITSFLAG == 4)                                   // Shelly an oder aus
  {
    while (T_MODE == 1)
    {
      LCDZ2();
      if ((T_DOWN == 0)&&(SHELLY == 1))
      SHELLY = 0;
      if ((T_UP == 0)&&(SHELLY == 0))
      SHELLY = 1;
      LCDZ2();
      if (SHELLY == 0)
      LCDText("Shelly aus      ");
      else 
      LCDText("Shelly an       ");
      Delaymittel();
    }
    while (T_MODE == 0)
    Delaymittel();
    CCPR1L = PWM_MIN;
  }

  while (T_MODE == 0)                                     // und alles in EEprom rein
  Delaymittel();
  EEDATA = U_MIN>>8 & 0b0000000011111111;
  EEADR = 1;
  EESchreibe();
  EEDATA = U_MIN & 0b0000000011111111;
  EEADR = 2;
  EESchreibe();
  EEDATA = P_SOLL>>8 & 0b0000000011111111;
  EEADR = 3;
  EESchreibe();
  EEDATA = P_SOLL & 0b0000000011111111;
  EEADR = 4;
  EESchreibe();
  EEDATA = LADEFLAG;
  EEADR = 5;
  EESchreibe();
  EEDATA = SHELLY;
  EEADR = 6;
  EESchreibe();
  EEDATA = PWM_MIN;
  EEADR = 7;
  EESchreibe();
  EEDATA = ARBEITSFLAG;
  EEADR = 8;
  EESchreibe();
  LCDLeer();
  LCDZ1();
  LCDText("Gespeichert     ");
}


void Init (void)                                          // Initialisierungsroutine
{ 
  PORTA = 0;
  PORTB = 0;
  ANSELA = 0b00010011;
  ANSELB = 0b00000000;
  ADCON1 = 0b10110011;                                    // Links justiert, 4V Referenz, int. Generator
  FVRCON = 0b11000011;
  ADCON0 = 0b00000001;
  TRISA = 0b00111111;      
  TRISB = 0b00000110;
  OSCCON = 0b11110000;                                    //32 MHz. int. Oszillator 
  APFCON0 = 0b00000001;
  CCP1CON = 0b00001100;			                          // PWM Mode	
  PR2 = 0b11111111;
  MDCON = 0b00000000;
  T2CON = 0b00000100;			                          // Timer2 ON, kein Pre- kein Post-counter
 
  EEADR = 1;
  EELese();
  U_MIN = EEDATA;
  U_MIN = U_MIN<<8;
  EEADR = 2;
  EELese();
  U_MIN = U_MIN | EEDATA;
  if (U_MIN > 600)
  U_MIN = 550;
  EEADR = 3;
  EELese();
  P_SOLL = EEDATA;
  P_SOLL = P_SOLL<<8;
  EEADR = 4;
  EELese();
  P_SOLL = P_SOLL | EEDATA;
  if (P_SOLL > 500)
  P_SOLL = 100;
  EEADR = 5;
  EELese();
  LADEFLAG = EEDATA;
  if (LADEFLAG > 1)
  LADEFLAG = 0;
  EEADR = 6;
  EELese();
  SHELLY = EEDATA;
  if (SHELLY > 1)
  SHELLY = 0;
  EEADR = 7 ;
  EELese();
  PWM_MIN = EEDATA;
  if (PWM_MIN > 120)
  PWM_MIN = 40;
  EEADR = 8;
  EELese();
  ARBEITSFLAG = EEDATA;
  if (ARBEITSFLAG > 4)
  ARBEITSFLAG = 1;
}                             


void EESchreibe (void)
{
  EECON1bits.EEPGD = 0;
  EECON1bits.CFGS = 0;
  EECON1bits.WREN = 1;
  INTCONbits.GIE = 0; 
  EECON2 = 0b01010101;
  EECON2 = 0b10101010;
  EECON1bits.WR = 1;
  while (EECON1bits.WR == 1);      // Warten, bis Schreiben fertig
  INTCONbits.GIE = 1; 
  EECON1bits.WREN = 0;
}


void Delaymittel(void)
{
  unsigned int i;
  for (i = 0; i < 30000; i++);
}


// Ab hier kommen nur noch die ganzen LCD Routinen und Delays, die das altertmliche 16x2 Display bedienen. 

void EELese (void)
{
  EECON1bits.EEPGD = 0;
  EECON1bits.CFGS = 0;
  EECON1bits.RD = 1;               // EEDATA = eingelesener Wert
}


void LCDAD1 ()               // konvertiert DATA nach TEMP1 und sendet auf LCD mit Komma.
{
  TEMP1 = DATA1;         
  if (TEMP1 >= 999)   // grer 999, Tausender Stellen
  {
    
    TEMP1 = (TEMP1 / 1000);
    TEMP1 = (TEMP1 + 48);
    LCDWrite(TEMP1);
  }
  else LCDText(" ");
  TEMP1 = DATA1;   
  while (TEMP1 >= 1000)
  TEMP1 = (TEMP1 - 1000);
  if (TEMP1 >= 99)   // grer 99, Hunderter Stellen
  {
    
    TEMP1 = (TEMP1 / 100);
    TEMP1 = (TEMP1 + 48);
    LCDWrite(TEMP1);
  }
  else LCDText("0");
  TEMP1 = DATA1;
  while (TEMP1 >= 100)
  TEMP1 = (TEMP1 - 100);
  if (TEMP1 > 9)   // Grer 9, Zehner Stellen
  {
    TEMP1 = (TEMP1 / 10);
    TEMP1 = (TEMP1 + 48);
    LCDWrite(TEMP1);
  }  
  else LCDText("0");
  LCDText(",");
  TEMP1 = DATA1;
  while (TEMP1 >= 100)
  TEMP1 = (TEMP1 -100);
  while (TEMP1 >= 10)
  TEMP1 = (TEMP1 -10);
  TEMP1 = (TEMP1 + 48);
  LCDWrite (TEMP1);
}


void LCDAD2 ()               // konvertiert DATA1 nach TEMP1 und sendet auf LCD.
{
  TEMP1 = DATA1;
  while (TEMP1 >= 1000)
  TEMP1 = (TEMP1 - 1000);      
  if (TEMP1 > 99)   // grer 99, Hunderter Stellen
  {
    
    TEMP1 = (TEMP1 / 100);
    TEMP1 = (TEMP1 + 48);
    LCDWrite(TEMP1);
  }
  else LCDText(" ");
  TEMP1 = DATA1;
  while (TEMP1 >= 100)
  TEMP1 = (TEMP1 - 100);
  if (TEMP1 > 9)   // Grer 9, Zehner Stellen
  {
    TEMP1 = (TEMP1 / 10);
    TEMP1 = (TEMP1 + 48);
    LCDWrite(TEMP1);
  }  
  else LCDText("0");
  TEMP1 = DATA1;
  while (TEMP1 >= 100)
  TEMP1 = (TEMP1 -100);
  while (TEMP1 >= 10)
  TEMP1 = (TEMP1 -10);
  TEMP1 = (TEMP1 + 48);
  LCDWrite (TEMP1);
}


void LCDLeer (void)
{
  LCD_RS=0;                      // 1 = Daten, 0 = Befehl
  LCD_EN=0;
  if (RELAIS == 1)
  PORTB = 0b00001001;      //write upper nibble to port
  if (RELAIS == 0)
  PORTB = 0b00000001;
  LCDDelay500ns();
  LCD_EN = 1;                    // Clock the cmd in
  LCDDelay500ns();
  LCD_EN = 0;        
  if (RELAIS == 1)
  PORTB = 0b00011001;      
  if (RELAIS == 0)
  PORTB = 0b00010001;
  LCDDelay500ns();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms();
}


void LCDZ1 (void)
{
  LCD_RS=0;                      // 1 = Daten, 0 = Befehl
  LCD_EN=0;
  if (RELAIS == 1)
  PORTB = 0b10001000;            //write upper nibble to port
  if (RELAIS == 0)
  PORTB = 0b10000000;
  LCDDelay500ns();
  LCD_EN = 1;                    // Clock the cmd in
  LCDDelay500ns();
  LCD_EN = 0;    
  if (RELAIS == 1)    
  PORTB = 0b00001000;
  if (RELAIS == 0)
  PORTB = 0b00000000;
  LCDDelay500ns();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms();
}


void LCDZ2 (void)
{
  LCD_RS=0;                      // 1 = Daten, 0 = Befehl
  LCD_EN=0;
  if (RELAIS == 1)
  PORTB = 0b10101000;            //Shift and write
  if (RELAIS == 0)
  PORTB = 0b10100000;
  LCDDelay500ns();
  LCD_EN = 1;                    // Clock the cmd in
  LCDDelay500ns();
  LCD_EN = 0;   
  if (RELAIS == 1)     
  PORTB = 0b10001000;    
  if (RELAIS == 0)
  PORTB = 0b10000000; 
  LCDDelay500ns();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms();
}


void LCDBefehl(char DATA)
{
  LCD_RS=0;                      // 1 = Daten, 0 = Befehl
  LCD_EN=0;
  if (RELAIS == 1)
  PORTB = DATA|0x0f;             //write upper nibble to port
  if (RELAIS == 0)
  PORTB = DATA&0xf0;
  LCDDelay500ns();
  LCD_EN = 1;                    // Clock the cmd in
  LCDDelay500ns();
  LCD_EN = 0;       
  if (RELAIS == 1) 
  PORTB = ((DATA<<4)|0x0f);      //Write lower nibble to port
  if (RELAIS == 0)
  PORTB = ((DATA<<4)&0xf0);
  LCDDelay500ns();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms();
}


void LCDWrite (char TEMP)
{
  LCD_RS=1;                      // 1 = Daten, 0 = Befehl
  LCD_EN=0;
  if (RELAIS == 1)
  PORTB = TEMP|0x0f;             //Shift 4 times and write upper nibble to port
  if (RELAIS == 0)
  PORTB = TEMP&0xf0;
  LCDDelay500ns();
  LCD_EN = 1;                    // Clock the cmd in
  LCDDelay500ns();
  LCD_EN = 0;    
  if (RELAIS == 1)
  PORTB = ((TEMP<<4)|0x0f);	     //Write lower nibble to port
  if (RELAIS == 0)
  PORTB = ((TEMP<<4)&0xf0);
  LCDDelay500ns;
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms();
}


void LCDText (const char *buffer)
{
  while (*buffer)
  {
    LCDWrite(*buffer);
    *buffer++;
  } 
}


void LCDInit (void)
{
  LCDDelay15ms();
  LCDDelay15ms();
  PORTB = 0b00110000;
  LCDDelay15ms ();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms ();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms ();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  PORTB = 0b00100000;
  LCDDelay4ms ();
  LCD_EN = 1;
  LCDDelay500ns();
  LCD_EN = 0;
  LCDDelay4ms ();
  LCDBefehl(0b00101000);
  LCDBefehl(0b00000001);
  LCDBefehl(0b00000110);
  LCDBefehl(0b00001100);
}


void LCDDelay15ms (void)
{
  int i;
  for(i=0;i<2000;i++)
  {
    NOP();
  }
  return;
}


void LCDDelay4ms (void)
{
  int i;
  for(i=0;i<500;i++)
  {
    NOP();
  }
  return;
}


void LCDDelay500ns(void)
{
  NOP();
  NOP();
}
